#version 130
#extension GL_EXT_geometry_shader4 : enable

uniform sampler2D adjacency; 
uniform sampler2D shortestPath; //shortest path length + path dir
out vec4 nodedata;

void main(void)
{
    float nodeCount = textureSize(shortestPath, 0).x;
    vec4 vertexdata = gl_PositionIn[0];
    int nodeID = int(vertexdata.x);

    vec4 storedNodeData = texelFetch(shortestPath, ivec2(nodeID, 0), 0);
    float pathLen = storedNodeData.y;

    //this should be the start node
    if(vertexdata.w == 0)
    {
         gl_Position = vec4(float(nodeID) / nodeCount * 2.0 - 1.0, 0, 0, 1);
         nodedata = vec4(vertexdata.xyz, 1);
         EmitVertex();
    }
    
    //visit all neighbor nodes
    for(int i = 0; i < nodeCount; ++i)
    {
     float edgeWeight = texelFetch(adjacency, ivec2(nodeID, i), 0).r;
     if(edgeWeight > 0)
     {
         vec4 neighNodeData = texelFetch(shortestPath, ivec2(i, 0), 0);
         float neighPathLen = neighNodeData.y;
         float neighPathLenNew = pathLen + edgeWeight;
         if(neighPathLenNew < neighPathLen || (neighPathLen == 0 && neighNodeData.w == 0))
         {
            gl_Position = vec4(float(i) / nodeCount * 2.0 - 1.0, 0, 0, 1);
            nodedata = vec4(i, neighPathLenNew, nodeID, 1);
            EmitVertex();
         }
     }
    }
    
}