#include <GL/glew.h>
#include <GL/freeglut.h>

#include <iostream>

#include "framebuffer.hpp"
#include "shader.hpp"
#include "texture.hpp"
#include "quad.hpp"
#include "GraphRenderer.h"

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

int windowWidth = 600;
int windowHeight = 600;

int nodeCnt = 8;
float* data;

Framebuffer* nodes;
GLuint adjacency;
Quad* fullscreenQuad;
Shader* exploreShader;
Shader* pathShader;

GLuint streamoutBufffer[2];
GLuint outputQuery;


int startNode = 0;
int endNode = 1;

GraphRenderer* graphrenderer;

bool startChanged = true;
bool endChanged = true;

void init()
{
    glewExperimental = GL_TRUE;
    GLenum err = glewInit();
    if (GLEW_OK != err)	{
        std::cerr << "Error: "<< glewGetErrorString(err) << std::endl;
    } else {
        if (GLEW_VERSION_3_3)
        {
            std::cout<< "Driver supports OpenGL 3.3\nDetails:"<< std::endl;
            std::cout << "Using GLEW " << glewGetString(GLEW_VERSION)<< std::endl;
            std::cout << "Vendor: " << glGetString (GL_VENDOR)<< std::endl;
            std::cout << "Renderer: " << glGetString (GL_RENDERER)<< std::endl;
            std::cout << "Version: " << glGetString (GL_VERSION)<< std::endl;
            std::cout << "GLSL: " << glGetString (GL_SHADING_LANGUAGE_VERSION)<< std::endl;
        }
    }
    glGetError();

    fullscreenQuad = new Quad();
    exploreShader = new Shader("explore.vert", "explore.frag", "explore.geom");

    nodes = new Framebuffer(nodeCnt, 1, 1, false, true);		
    
    glGenTextures(1, &adjacency);
    data = new float[nodeCnt*nodeCnt];
    for(int i = 0; i < nodeCnt; ++i){
        for(int j = 0; j < nodeCnt; ++j)
        {
            if(i == j)
                data[i* nodeCnt + j] = 0;
            else
                data[i* nodeCnt + j] = (float)rand() / (float)(RAND_MAX);
        }
    }
    glBindTexture(GL_TEXTURE_2D, adjacency);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, nodeCnt, nodeCnt, 0, GL_RED, GL_FLOAT, data);

    graphrenderer = new GraphRenderer(data, nodeCnt);

    glGenBuffers(2,streamoutBufffer);
    glBindBuffer(GL_ARRAY_BUFFER_ARB,streamoutBufffer[0]);
    glBufferData(GL_ARRAY_BUFFER_ARB,nodeCnt*nodeCnt*sizeof(float)*2,0,GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER_ARB,streamoutBufffer[1]);
    glBufferData(GL_ARRAY_BUFFER_ARB,nodeCnt*nodeCnt*sizeof(float)*2,0,GL_DYNAMIC_DRAW);
    glGenQueries(1, &outputQuery);
}

void draw()
{
    if(startChanged)
    {
        startChanged = false;
        nodes->setRenderTarget();
        glClearColor(0,0,0,0);
        glClear(GL_COLOR_BUFFER_BIT);

        exploreShader->enable();

        exploreShader->bindUniformTexture("adjacency", adjacency, 0);
        exploreShader->bindUniformTexture("shortestPath", nodes->getColorBuffer(0), 1);

        glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER,0,streamoutBufffer[0]);

        glPointSize(1);

        glBeginTransformFeedback(GL_POINTS); 
        glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN, outputQuery);
        
        glBegin(GL_POINTS);
        glVertex4f(startNode, 0, 0, 0);
        glEnd();
        glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN);

        glEndTransformFeedback();

        GLint outPointCount = 0;
        GLint succ = 0;

        while(!succ)
            glGetQueryObjectiv(outputQuery, GL_QUERY_RESULT_AVAILABLE, &succ); 
        glGetQueryObjectiv(outputQuery, GL_QUERY_RESULT, &outPointCount);
        succ = 0;

        std::cout << "points written: " << outPointCount << std::endl;
        glEnableClientState(GL_VERTEX_ARRAY);			
        int bb = 0;
        while(outPointCount > 0)
        {			
            glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER,0,streamoutBufffer[(bb+1)%2]);
            glBeginTransformFeedback(GL_POINTS); 

            glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN, outputQuery);

            glBindBufferARB(GL_ARRAY_BUFFER_ARB, streamoutBufffer[bb]);
            glVertexPointer(4,GL_FLOAT,0,NULL);
            glDrawArrays(GL_POINTS, 0, outPointCount);			

            glEndTransformFeedback();

            glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN);
            while(!succ)
                glGetQueryObjectiv(outputQuery, GL_QUERY_RESULT_AVAILABLE, &succ); 
            glGetQueryObjectiv(outputQuery, GL_QUERY_RESULT, &outPointCount);
            succ = 0;

            std::cout << "points written: " << outPointCount << std::endl;

            bb = (bb + 1) % 2;	
        }		
        glDisableClientState(GL_VERTEX_ARRAY); 			
        exploreShader->disable();		

        nodes->disableRenderTarget();

    }
    
    glViewport(0, 0, windowWidth, windowHeight);

    glClearColor(0.5,0.5,0.5,1);
    glClear(GL_COLOR_BUFFER_BIT);

    graphrenderer->render();

    //draw start and end points
    glBegin(GL_POINTS);
	glColor3f(0,0,1);
    float x0,y0,x1,y1;
    graphrenderer->getNodePos(startNode,x0,y0);
    glVertex3f(x0,y0,0);
	glColor3f(0,1,0);
    graphrenderer->getNodePos(endNode,x1,y1);
	glVertex3f(x1,y1,0);
	glEnd();

    //draw path
    static float* pathData = new float[nodeCnt*4];
    glBindTexture(GL_TEXTURE_2D, nodes->getColorBuffer(0));
    glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_FLOAT,pathData);

    glColor3f(1,0,0);
    glBegin(GL_LINE_STRIP);
    float x, y;
    int node = endNode;
    while(node != startNode)
    {
        graphrenderer->getNodePos(node, x, y);
        glVertex3f(x,y,0);
        node = pathData[node*4 + 2];
        printf("newnode: %i \n", node);
    }
    graphrenderer->getNodePos(node, x, y);
    glVertex3f(x,y,0);
    glEnd();
    
    glFinish();
}

void mouse(int button, int state, int x, int y)
{
    if(GLUT_LEFT_BUTTON == button && GLUT_UP == state)
    {	
        startNode = graphrenderer->getClosest((float) x / (float) windowWidth * 2.0f -1.0f, 
                                              -1 * ((float) y / (float) windowHeight * 2.0f -1.0f));
        startChanged = true;
        glutPostRedisplay();
    }
    if(GLUT_RIGHT_BUTTON == button && GLUT_UP == state)
    {	
        endNode = graphrenderer->getClosest((float) x / (float) windowWidth * 2.0f -1.0f, 
                                              -1 * ((float) y / (float) windowHeight * 2.0f -1.0f));
        glutPostRedisplay();
    }
}


int main(int argc, char* argv[]){
    glutInit(&argc, argv);
    glutInitContextVersion (3, 0);
    glutInitContextFlags (GLUT_CORE_PROFILE | GLUT_DEBUG);
    glutInitDisplayMode(GLUT_RGBA | GLUT_ALPHA );
    glutInitWindowSize(windowWidth, windowHeight);
    glutCreateWindow("GPGPU 3. labor: glsl");

    init();

    glutDisplayFunc(draw);
    glutMouseFunc(mouse);
    glutMainLoop();

    return(0);
}
