#include <GL/glew.h>
#include <GL/freeglut.h>

#include <iostream>

#include "framebuffer.hpp"
#include "shader.hpp"
#include "texture.hpp"
#include "quad.hpp"

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

int windowWidth = 600;
int windowHeight = 600;

int dataSize = 1024;

Quad* fullscreenQuad;
Shader* compareShader;

Framebuffer* buffer;


void init(){

	glewExperimental = GL_TRUE;
	GLenum err = glewInit();
	if (GLEW_OK != err)	{
		std::cerr << "Error: "<< glewGetErrorString(err) << std::endl;
	} else {
		if (GLEW_VERSION_3_3)
		{
			std::cout<< "Driver supports OpenGL 3.3\nDetails:"<< std::endl;
			std::cout << "Using GLEW " << glewGetString(GLEW_VERSION)<< std::endl;
			std::cout << "Vendor: " << glGetString (GL_VENDOR)<< std::endl;
			std::cout << "Renderer: " << glGetString (GL_RENDERER)<< std::endl;
			std::cout << "Version: " << glGetString (GL_VERSION)<< std::endl;
			std::cout << "GLSL: " << glGetString (GL_SHADING_LANGUAGE_VERSION)<< std::endl;
		}
	}
	glGetError();


	fullscreenQuad = new Quad();
	compareShader = new Shader("compute.vert", "compute.frag");

    buffer = new Framebuffer(dataSize, dataSize, 1, true, true);		

    float* data = new float[dataSize*dataSize*4];
    float mini = 2;
    for(int i = 0; i < dataSize*dataSize*4; ++i){
        data[i] = (float)rand() / (float)(RAND_MAX) + (float)rand() / (float)(RAND_MAX);
        mini = mini < data[i] ?  mini : data[i];
    }
    printf("Minimum: %f\n", mini);

    glBindTexture(GL_TEXTURE_2D, buffer->getColorBuffer(0));
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, dataSize, dataSize, 0, GL_RGBA, GL_FLOAT, data);

    delete data;
	
}

void compute(){

    glClearColor(0,1,0,1);
    glClear(GL_COLOR_BUFFER_BIT);
    int numMipLevels = buffer->getLevels();
    compareShader->enable();        
	for(int l = 1; l < numMipLevels; l++)
	{
        buffer->setRenderTarget(l);

        glClearColor(0,1,0,1);
        glClear(GL_COLOR_BUFFER_BIT);

        compareShader->bindUniformInt("level", l);
        compareShader->bindUniformTexture("data", buffer->getColorBuffer(0), 0);
        fullscreenQuad->render(compareShader);
        
	}
    compareShader->disable();
    glFinish();
    float result[4];
    buffer->disableRenderTarget();
    glBindTexture(GL_TEXTURE_2D, buffer->getColorBuffer(0));
    glGetTexImage(GL_TEXTURE_2D,  buffer->getLevels()-1, GL_RGBA, GL_FLOAT, result);
    printf("Result: %f %f %f %f\n", result[0],result[1],result[2],result[3]);
}

int main(int argc, char* argv[]){
	glutInit(&argc, argv);
	glutInitContextVersion (3, 0);
	glutInitContextFlags (GLUT_CORE_PROFILE | GLUT_DEBUG);
	glutInitDisplayMode(GLUT_RGBA );
	glutInitWindowSize(windowWidth, windowHeight);
	glutCreateWindow("GPGPU 3. labor: glsl");
    glutHideWindow();

	init();
    compute();

	return(0);
}
